#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Nov 29, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : metrics.py
project  : nadl
license  : GPL-3.0+

* Metric class and functions
* Some simple metrics that not in optax loss.  For complex metrics, use sklearn.
"""

# ruff: noqa: F722
from functools import partial
from warnings import catch_warnings, simplefilter, warn

import jax
import jax.numpy as jnp
from equinox import (
  AbstractVar,
  Module,
  combine,
  filter_pure_callback,
  is_array,
  partition,
  tree_equal,
  tree_pformat,
)
from optax import softmax_cross_entropy_with_integer_labels

from jaxtyping import Array, ArrayLike, Float, Int, Num, Scalar
from typing import Literal, Self, cast

import numpy as np
import sklearn.metrics as m

from .utils import batch_array_p, filter_concat, filter_tree


def convert(x: Num[ArrayLike, "..."]) -> Num[Array, "B ..."]:
  """Convert to float."""
  x = jnp.asarray(x)
  return x.reshape(-1, *x.shape[1:])


def roc_auc_score(labels: Num[Array, " A"], preds: Num[Array, " A"]) -> Scalar:
  """Compute ROC."""
  return jax.pure_callback(
    m.roc_auc_score, jax.ShapeDtypeStruct((), jnp.float32), labels, preds
  )


def average_precision_score(
  labels: Num[Array, " A"],
  preds: Num[Array, " A"],
  average: Literal["micro", "macro"] = "macro",
) -> Scalar:
  """Compute PR."""
  return jax.pure_callback(
    partial(m.average_precision_score, average=average),
    jax.ShapeDtypeStruct((), jnp.float32),
    labels,
    preds,
  )


def pr_auc_score(labels: Num[Array, " A"], preds: Num[Array, " A"]) -> Scalar:
  """Compute PR."""

  def _callback(lbl: Num[Array, " A"], prd: Num[Array, " A"]) -> Scalar:
    precision, recall, _ = m.precision_recall_curve(lbl, prd)
    return jnp.asarray(m.auc(recall, precision))

  return filter_pure_callback(
    _callback,
    labels,
    preds,
    result_shape_dtypes=jax.ShapeDtypeStruct((), jnp.float32),
  )


class Metric[**P, T](Module):
  """Base Metric.

  This metric is for batch data not for a single one.
  Thus, please consider the shape of the input data as [b, ...].

  The default getitem will only take the array with ndim > 1.
  """

  name: AbstractVar[str | None]

  def __post_init__(self) -> None:
    """Post init."""
    for k, v in self.__dict__.items():
      if isinstance(v, Array | np.ndarray):
        self.__dict__[k] = convert(v)

  def __check_init__(self) -> None:  # noqa: PLW3201
    """Check init."""
    arrs = filter_tree(self, batch_array_p)
    if arrs:
      if not tree_equal(*jax.tree.map(lambda x: jnp.shape(x)[0], arrs)):
        raise ValueError("All batched array should have the same batch size.")
    else:
      warn("No batched array found in the metric.", stacklevel=1)

  @classmethod
  def merge(cls, *metrics: Self) -> Self:
    """Merge all metrics."""
    _, s2 = partition(metrics, is_array)
    if not tree_equal(*s2):
      raise ValueError("All metrics should have the same non-array values.")
    return filter_concat(metrics, batch_array_p)

  def __add__(self, value: Self) -> Self:
    """Add."""
    return self.merge(self, value)

  def __or__(self, value: Self) -> Self:
    """Or."""
    return combine(self, value)

  @classmethod
  def empty(cls, *_args: P.args, **_kwds: P.kwargs) -> Self:
    """Empty."""
    with catch_warnings():
      simplefilter("ignore")
      return cls(**dict.fromkeys(cls.__dataclass_fields__))

  def compute(self) -> T:
    """Compute."""
    raise NotImplementedError

  def __getitem__(self, idx: int | slice | Int[ArrayLike, "..."]) -> Self:
    """Get item.

    The default getitem will only take the array with ndim > 1.
    """
    return jax.tree.map(lambda x: x[idx] if batch_array_p(x) else x, self)

  def show(self) -> str:
    """Show."""
    return tree_pformat(self, short_arrays=False)


class Accuracy(Metric):
  """Accuracy."""

  labels: Int[Array, "..."]
  preds: Int[Array, "..."]
  name: str = "accuracy"

  def compute(self) -> Array:
    """Compute."""
    return jnp.nanmean(self.labels == self.preds, axis=-1)


def dice_coef(
  y_true: Int[Array, " A"], y_pred: Int[Array, " A"], eps: float = 1e-8
) -> Scalar:
  """Compute dice coefficient."""
  y_true = jnp.asarray(y_true)
  y_pred = jnp.asarray(y_pred)

  intersection = jnp.sum(y_true * y_pred)
  union = jnp.sum(y_true) + jnp.sum(y_pred)

  return (2.0 * intersection) / (union + eps)


def iou_coef(
  y_true: Int[Array, " A"], y_pred: Int[Array, " A"], eps: float = 1e-8
) -> Scalar:
  """Compute intersection over union."""
  y_true = jnp.asarray(y_true)
  y_pred = jnp.asarray(y_pred)

  intersection = jnp.sum(y_true * y_pred)
  union = jnp.sum(y_true) + jnp.sum(y_pred) - intersection

  return intersection / (union + eps)


def info_nce(
  pos: Float[Array, "B D"], neg: Float[Array, "B D"], t: float = 0.07
) -> Float[Array, "B 1"]:
  """Compute info nce loss for paired pos-neg data."""
  pos /= jnp.linalg.norm(pos, axis=-1, keepdims=True)
  neg /= jnp.linalg.norm(neg, axis=-1, keepdims=True)

  pos_sim = jnp.einsum("bi,bi->b", pos, pos) / t
  neg_sim = jnp.einsum("bi,nd->bn", pos, neg) / t
  logits = jnp.c_[pos_sim, neg_sim]
  return cast(
    Float[Array, "B 1"],
    softmax_cross_entropy_with_integer_labels(logits, jnp.arange(logits.shape[0])),
  )


def _test() -> None:
  """Test."""
  k1, k2 = jax.random.split(jax.random.key(42))
  p = jax.random.normal(k1, (4, 30))
  n = jax.random.normal(k2, (4, 30))
  assert info_nce(p, n).shape == (4,)


if __name__ == "__main__":
  _test()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Dec  8, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : blocks.py
project  : nadl
license  : GPL-3.0+

Layer Blocks
"""

from collections.abc import Callable

import equinox as eqx
import jax
import jax.numpy as jnp

from jaxtyping import Array, Float, PRNGKeyArray


class SplineLinear(eqx.Module):
  """Spline Linear layer."""

  init_scale: float
  liner: eqx.nn.Linear

  def __init__(
    self,
    in_features: int,
    out_features: int,
    init_scale: float = 1.0,
    use_bias: bool = True,
    *,
    key: PRNGKeyArray,
  ) -> None:
    """Initialize."""
    self.init_scale = init_scale
    linear = eqx.nn.Linear(in_features, out_features, use_bias, key=key)
    self.linear = eqx.tree_at(
      lambda ly: ly.weight,
      linear,
      jax.nn.initializers.truncated_normal(stddev=init_scale)(key, linear.weight.shape),
    )

  def __call__(self, x: Float[Array, " A"]) -> Float[Array, " A"]:
    """Forward."""
    return self.linear(x)


class RadialBasisFunction(eqx.Module):
  """Gaussian Radial Basis Function."""

  grid: Float[Array, " A"]
  num_grids: int
  grid_min: float
  grid_max: float
  denominator: float

  def __init__(
    self,
    grid_min: float = -2.0,
    grid_max: float = 2.0,
    num_grids: int = 5,
    denominator: float = 1,  # larger denominators lead to smoother basis
  ) -> None:
    """Initialize."""
    self.grid = jnp.linspace(grid_min, grid_max, num_grids)
    self.num_grids = num_grids
    self.grid_min = grid_min
    self.grid_max = grid_max
    self.denominator = denominator or (grid_max - grid_min) / (num_grids - 1)

  def __call__(self, x: Float[Array, " A"]) -> Float[Array, " A"]:
    """Forward."""
    return jnp.exp(-(((x[..., None] - self.grid) / self.denominator) ** 2))


class FastKANLayer(eqx.Module):
  """Gaussian Radial Basis Functions replace the spline fast KAN.

  From https://github.com/ZiyaoLi/fast-kan/blob/master/fastkan/fastkan.py
  """

  rbf: RadialBasisFunction
  spline_linear: SplineLinear
  use_base_update: bool
  use_layernorm: bool
  layernorm: eqx.nn.LayerNorm | None = None
  base_activation: Callable[[Float[Array, "..."]], Float[Array, "..."]] = jax.nn.silu
  base_linear: eqx.nn.Linear | None = None

  def __init__(  # noqa: PLR0913
    self,
    input_dim: int,
    output_dim: int,
    grid_min: float = -2.0,
    grid_max: float = 2.0,
    num_grids: int = 5,
    use_base_update: bool = True,
    use_layernorm: bool = True,
    base_activation: Callable[[Float[Array, "..."]], Float[Array, "..."]] = jax.nn.silu,
    spline_weight_init_scale: float = 0.1,
    *,
    key: PRNGKeyArray,
  ) -> None:
    """Initialize."""
    self.use_layernorm = use_layernorm
    if use_layernorm:
      assert (
        input_dim > 1
      ), "Do not use layernorms on 1D inputs. Set `use_layernorm=False`."
      self.layernorm = eqx.nn.LayerNorm(input_dim)

    self.rbf = RadialBasisFunction(grid_min, grid_max, num_grids)

    self.spline_linear = SplineLinear(
      input_dim * num_grids, output_dim, spline_weight_init_scale, key=key
    )

    self.use_base_update = use_base_update
    if use_base_update:
      self.base_activation = base_activation
      self.base_linear = eqx.nn.Linear(input_dim, output_dim, key=key)

  def __call__(
    self, x: Float[Array, " A"]
  ) -> Float[Array, " A"]:
    """Forward."""
    if self.use_layernorm:
      assert self.layernorm is not None, "Layernorm is not initialized."
      x = self.layernorm(x)
    spline_basis = self.rbf(x)

    ret = self.spline_linear(spline_basis.reshape(*spline_basis.shape[:-2], -1))

    if self.use_base_update:
      assert self.base_linear is not None, "Base linear is not initialized."
      base = self.base_linear(self.base_activation(x))
      ret += base

    return ret

  @property
  def input_dim(self) -> int:
    """Get input dimension."""
    return self.spline_linear.linear.weight.shape[1] // self.rbf.num_grids

  @property
  def output_dim(self) -> int:
    """Get output dimension."""
    return self.spline_linear.linear.weight.shape[0]

  def plot_curve(
    self,
    input_index: int,
    output_index: int,
    num_pts: int = 1000,
    num_extrapolate_bins: int = 2,
  ) -> tuple[Float[Array, " A"], Float[Array, " A"]]:
    """This function returns the learned curves in a FastKANLayer.

    input_index: the selected index of the input, in [0, input_dim) .
    output_index: the selected index of the output, in [0, output_dim) .
    num_pts: num of points sampled for the curve.
    num_extrapolate_bins (N_e): num of bins extrapolating from the given grids. The
        curve will be calculate in the range of
        [grid_min - h * N_e, grid_max + h * N_e].
    """  # noqa: D404
    ng = self.rbf.num_grids
    h = self.rbf.denominator
    assert input_index < self.input_dim
    assert output_index < self.output_dim
    w = self.spline_linear.linear.weight[
      output_index, input_index * ng : (input_index + 1) * ng
    ]  # num_grids,
    x = jnp.linspace(
      self.rbf.grid_min - num_extrapolate_bins * h,
      self.rbf.grid_max + num_extrapolate_bins * h,
      num_pts,
    )  # num_pts, num_grids
    y = (w * self.rbf(x)).sum(-1)
    return x, y



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : May 14, 2024
email    : Nasy <nasyxx+python@gmail.com>
filename : states.py
project  : nadl
license  : GPL-3.0+

States
"""

from __future__ import annotations

import shutil
from abc import abstractmethod
from collections.abc import Callable

import equinox as eqx
import optax
import orbax.checkpoint as ocp

from jaxtyping import PyTree
from typing import TYPE_CHECKING, cast, Self

if TYPE_CHECKING:
  from pathlib import Path

  import jax

  from rich.console import Console


class BaseTrainState[T, M](eqx.Module):
  """Train state."""

  model: M
  tx: optax.GradientTransformation
  opt_state: optax.OptState
  loss: jax.Array
  step: jax.Array
  conf: T

  @classmethod
  @abstractmethod
  def init[**P](cls: type[Self], *args: P.args, **kwds: P.kwargs) -> Self:
    """Create initial state."""
    raise NotImplementedError

  def apply_grads(self, loss: jax.Array, grads: eqx.Module) -> BaseTrainState[T, M]:
    """Apply gradients."""
    updates, opt_state = self.tx.update(
      cast(optax.Updates, grads), self.opt_state, params=cast(optax.Params, self.model)
    )
    model = eqx.apply_updates(self.model, updates)
    return eqx.tree_at(
      lambda x: (x.model, x.opt_state, x.loss, x.step),
      self,
      (model, opt_state, loss, self.step + 1),
    )


type T_savefn = Callable[[int, BaseTrainState, PyTree, PyTree], None]


def state_fn(
  rpath: Path,
  console: Console | None = None,
  keeps: int = 5,
  clean: bool = False,
  item_names: tuple[str, ...] | None = None,
  item_handlers: dict | None = None,
  best_fn: Callable[[PyTree], float] | None = None,
) -> tuple[ocp.CheckpointManager, T_savefn]:
  """Get states manager."""
  match (item_names, item_handlers):
    case (None, None):
      item_names = ("state", "extra_metadata")
      item_handlers = {
        "state": ocp.PyTreeCheckpointHandler(),
        "extra_metadata": ocp.PyTreeCheckpointHandler(),
      }
    case _ if item_names and item_handlers:
      for i in item_names:
        assert i in item_handlers, f"Item {i} not in item_handlers."
    case _:
      raise ValueError("item_names and item_handlers should be both None or not None.")

  if console:
    console.log(f"Checkpoint path at {rpath}")
  if rpath.exists() and clean:
    if console:
      console.log("Cleaning up checkpoint...")
    shutil.rmtree(rpath)

  mngr = ocp.CheckpointManager(
    rpath,
    options=ocp.CheckpointManagerOptions(
      max_to_keep=keeps, save_interval_steps=1, best_fn=best_fn
    ),
    item_names=item_names,
    item_handlers=item_handlers,
  )

  def save(
    step: int, state: BaseTrainState, metadata: PyTree = None, metrics: PyTree = None
  ) -> None:
    """Save state."""
    mngr.save(
      step,
      args=ocp.args.Composite(
        state=ocp.args.PyTreeSave(state),  # type: ignore
        extra_metadata=ocp.args.PyTreeSave(metadata or {}),  # type: ignore
      ),
      metrics=metrics or {},
    )

  return mngr, save



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Dec  8, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : nets.py
project  : nadl
license  : GPL-3.0+

Pre define networks.
"""

from collections.abc import Callable

import equinox as eqx
import jax

from jaxtyping import Array, Float, PRNGKeyArray

from .blocks import FastKANLayer


class pMTnet(eqx.Module):  # noqa: N801
  """pMTnet.

  https://github.com/tianshilu/pMTnet

  Deep learning neural network prediction tcr binding specificity to
  peptide and HLA based on peptide sequences. Please refer to our
  paper for more details: 'Deep learning-based prediction of T cell
  receptor-antigen binding specificity.

  (https://www.nature.com/articles/s42256-021-00383-2)

  Lu, T., Zhang, Z., Zhu, J. et al. 2021.
  """

  layers: eqx.nn.Sequential

  def __init__(
    self,
    inp: int,
    out: int = 1,
    hiddens: tuple[int, int, int] = (300, 200, 100),
    dropout_rate: float = 0.2,
    *,
    key: PRNGKeyArray,
  ) -> None:
    """Initialize the pMTnet."""
    k1, k2, k3, k4 = jax.random.split(key, 4)
    self.layers = eqx.nn.Sequential([
      eqx.nn.Linear(inp, hiddens[0], key=k1),
      eqx.nn.Dropout(p=dropout_rate),
      eqx.nn.Lambda(jax.nn.relu),
      eqx.nn.Linear(hiddens[0], hiddens[1], key=k2),
      eqx.nn.Lambda(jax.nn.relu),
      eqx.nn.Linear(hiddens[1], hiddens[2], key=k3),
      eqx.nn.Lambda(jax.nn.relu),
      eqx.nn.Linear(hiddens[2], out, key=k4),
    ])

  def __call__(self, x: Float[Array, " A"]) -> Float[Array, " A"]:
    """Forward."""
    return self.layers(x)


class FastKAN(eqx.Module):
  """FastKAN.

  FastKAN: Very Fast Implementation (Approximation) of Kolmogorov-Arnold Network.
  """

  layers: eqx.nn.Sequential

  def __init__(
    self,
    layers_hidden: list[int],
    grid_min: float = -2.0,
    grid_max: float = 2.0,
    num_grids: int = 8,
    use_base_update: bool = True,
    base_activation: Callable[[Float[Array, "..."]], Float[Array, "..."]] = jax.nn.silu,
    spline_weight_init_scale: float = 0.1,
    *,
    key: PRNGKeyArray,
  ) -> None:
    """Initialize the FastKAN."""
    ks = jax.random.split(key, len(layers_hidden) - 1)
    self.layers = eqx.nn.Sequential([
      FastKANLayer(
        in_dim,
        out_dim,
        grid_min=grid_min,
        grid_max=grid_max,
        num_grids=num_grids,
        use_base_update=use_base_update,
        base_activation=base_activation,
        spline_weight_init_scale=spline_weight_init_scale,
        key=k,
      )
      for in_dim, out_dim, k in zip(
        layers_hidden[:-1], layers_hidden[1:], ks, strict=True
      )
    ])

  def __call__(self, x: Float[Array, " A"]) -> Float[Array, " A"]:
    """Forward."""
    return self.layers(x)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Jun 17, 2024
email    : Nasy <nasyxx+python@gmail.com>
filename : surgery.py
project  : nadl
license  : GPL-3.0+

Surgery for NADL.
"""

from typing import Any, Callable, TypeGuard
import jax
import jax.numpy as jnp
import equinox as eqx
from .typings import F, K


def init_fn(fn: Callable[[K, tuple[int, ...]]]) -> Callable[[F, K], F]:
  """Initialize function."""

  def _init_fn(weight: F, key: K) -> F:
    return fn(key, jnp.shape(weight))

  return _init_fn


def kaiming_init(weight: F, key: K) -> F:
  """Kaiming initialization."""
  return init_fn(jax.nn.initializers.he_normal())(weight, key)


def is_linear(x: Any) -> TypeGuard[eqx.nn.Linear]:  # noqa: ANN401
  """Check if a module is a linear layer."""
  return isinstance(x, eqx.nn.Linear)


def init_surgery[T, M](
  model: T,
  key: K,
  is_leaf: Callable[[M], bool] = is_linear,
  init: Callable[[F, K], F] = kaiming_init,
) -> T:
  """Initialize model."""

  def _get_weights(m: Any) -> list[F]:  # noqa: ANN401
    return list(
      map(
        lambda x: x.weight,
        filter(is_leaf, jax.tree_util.tree_leaves(m, is_leaf=is_leaf)),
      )
    )

  weights = _get_weights(model)
  new_weights = list(map(init, weights, jax.random.split(key, len(weights))))
  return eqx.tree_at(_get_weights, model, new_weights)
